
#include "cnf.h" 


using namespace Cnf;


CnfVariable::CnfVariable(int id, const std::string& name)
{
	this->value = CnfVariableValue::Undefined;
	this->name = name;
	this->id = id;
	this->numUnsolvedPositives = 0;
	this->numUnsolvedNegatives = 0;
}
CnfVariable::~CnfVariable()
{
}
void CnfVariable::Write(std::ostream& os, bool humanreadable)
{
	if(humanreadable)
	{
		if(name != "")
			os << name;
		else
			os << "#" << id;
	}
	else
		os << id;
}
void CnfVariable::PrintAssignment(std::ostream& os)
{
	if(name != "")
	{
		os << name << " = ";
		switch(value)
		{
			case CnfVariableValue::True:      os <<     "true"; break;
			case CnfVariableValue::False:     os <<    "false"; break;
			case CnfVariableValue::Undefined: os << "whatever"; break;
		}
		os << std::endl;
	}
	/*
	switch(value)
	{
		case CnfVariableValue::True:      os <<  ""; break;
		case CnfVariableValue::False:     os << "-"; break;
		case CnfVariableValue::Undefined: os << "?"; break;
	}
	os << id << " ";
	*/
}



void CnfVariable::SetValue(CnfVariableValue::CnfVariableValue newvalue)
{
	if(value == newvalue)
		return;
	CnfVariableValue::CnfVariableValue oldvalue = value;
	value = newvalue;
	for(CnfLiteral* l : positives)
		l->onVariableValueChanged(oldvalue, newvalue);
	for(CnfLiteral* l : negatives)
		l->onVariableValueChanged(oldvalue, newvalue);
}
void CnfVariable::RegisterLiteral(bool positive, CnfLiteral* literal)
{
	if(positive)
	{
		positives.push_back(literal);
		numUnsolvedPositives++;
	}
	else
	{
		negatives.push_back(literal);
		numUnsolvedNegatives++;
	}
}
void CnfVariable::onClauseSatisfactionChanged(CnfClause* clause, bool positiveLiteral, bool newSatisfaction)
{
	if(positiveLiteral)
		numUnsolvedPositives -= newSatisfaction ? 1 : -1;
	else
		numUnsolvedNegatives -= newSatisfaction ? 1 : -1;
}



CnfLiteral::CnfLiteral(bool positive, CnfVariable* variable)
{
	this->positive = positive;
	this->variable = variable;
	clause = nullptr;
	variable->RegisterLiteral(positive, this);
}
CnfLiteral::~CnfLiteral()
{
}
void CnfLiteral::onVariableValueChanged(CnfVariableValue::CnfVariableValue oldvalue, CnfVariableValue::CnfVariableValue newvalue)
{
	clause->onVariableValueChanged(variable, oldvalue, newvalue);
	bool wasSatisfied = ( positive && oldvalue == CnfVariableValue::True)
	                 || (!positive && oldvalue == CnfVariableValue::False);
	bool isSatisfied = ( positive && newvalue == CnfVariableValue::True)
	                || (!positive && newvalue == CnfVariableValue::False);
	if(wasSatisfied != isSatisfied)
		clause->onLiteralSatisfactionChanged(this, !wasSatisfied && isSatisfied);
}
void CnfLiteral::Write(std::ostream& os, bool humanreadable)
{
	os << (positive ? "" : (humanreadable ? "!" : "-"));
	variable->Write(os, humanreadable);
}
void CnfLiteral::onClauseSatisfactionChanged(CnfClause* clause, bool newSatisfaction)
{
	variable->onClauseSatisfactionChanged(clause, positive, newSatisfaction);
}



CnfClause::CnfClause(CnfFormula* parent)
{
	numSatisfiedLiterals = 0;
	numUnassignedLiterals = 0;
	formula = parent;
}
CnfClause::~CnfClause()
{
	for(CnfLiteral* l : literals)
		delete l;
}
CnfLiteral* CnfClause::addLiteral(bool positive, CnfVariable* variable)
{
	CnfLiteral* l = new CnfLiteral(positive, variable);
	l->clause = this;
	literals.push_back(l);
	numUnassignedLiterals++;
	return l;
}
void CnfClause::Write(std::ostream& os, bool humanreadable)
{
	for(CnfLiteral* l : literals)
	{
		l->Write(os, humanreadable);
		if(humanreadable)
		{
			if(l != *literals.rbegin())
				os << " || ";
		}
		else
			os << " ";
	}
}
void CnfClause::onVariableValueChanged(CnfVariable* variable, CnfVariableValue::CnfVariableValue oldvalue, CnfVariableValue::CnfVariableValue newvalue)
{
	bool wasDefined = oldvalue != CnfVariableValue::Undefined;
	bool isDefined = newvalue != CnfVariableValue::Undefined;
	if(wasDefined != isDefined)
		numUnassignedLiterals -= !wasDefined && isDefined ? 1 : -1;
}
void CnfClause::onLiteralSatisfactionChanged(CnfLiteral* literal, bool newSatisfaction)
{
	bool wasSatisfied = 0 < numSatisfiedLiterals;
	numSatisfiedLiterals += newSatisfaction ? 1 : -1;
	bool isSatisfied = 0 < numSatisfiedLiterals;
	
	if(wasSatisfied != isSatisfied)
	{
		formula->onClauseSatisfactionChanged(this, !wasSatisfied && isSatisfied);
		for(CnfLiteral* l : literals)
			l->onClauseSatisfactionChanged(this, !wasSatisfied && isSatisfied);
	}
}



CnfFormula::CnfFormula()
{
	numUnsolvedClauses = 0;
}
CnfFormula::~CnfFormula()
{
	for(CnfClause* c : clauses)
		delete c;
	for(CnfVariable* v : variables)
		delete v;
}
CnfVariable* CnfFormula::addVariable(std::string name)
{
	if(variableIndex.find(name) != variableIndex.end())
		return variableIndex[name];
	CnfVariable* v = new CnfVariable(1+variables.size(), name);
	variables.push_back(v);
	if(name != "")
		variableIndex[name] = v;
	return v;
}
CnfClause* CnfFormula::addClause(bool p1, CnfVariable* v1, bool p2, CnfVariable* v2, bool p3, CnfVariable* v3)
{
	CnfClause* c = new CnfClause(this);
	clauses.push_back(c);
	numUnsolvedClauses++;

	if(v1) c->addLiteral(p1, v1);
	if(v2) c->addLiteral(p2, v2);
	if(v3) c->addLiteral(p3, v3);
	return c;
}
void CnfFormula::Write(std::ostream& os, bool humanreadable)
{
	for(CnfClause* c : clauses)
	{
		c->Write(os, humanreadable);
		os << (humanreadable ? "" : " 0") << std::endl;
	}
	if(!humanreadable)
		os << "0" << std::endl;
}
void CnfFormula::onClauseSatisfactionChanged(CnfClause* clause, bool newSatisfaction)
{
	numUnsolvedClauses -= newSatisfaction ? 1 : -1;
}
void CnfFormula::PrintAssignment(std::ostream& os)
{
	for(CnfVariable* variable : variables)
		variable->PrintAssignment(os);
}
