

#include "cnf.h"
#include "sat.h"


using namespace Sat;


int SatFormula::nextId = 1;
SatFormula::SatFormula()
{
	id = nextId++;
}
SatFormula::~SatFormula()
{
}
void SatFormula::CollectVariables(std::set<std::string>* variables)
{
}
Cnf::CnfFormula* SatFormula::ToCnf()
{
	Cnf::CnfFormula* result = new Cnf::CnfFormula();
	Cnf::CnfVariable* tv = result->addVariable("\\true");
	result->addClause(true, tv);
	Cnf::CnfVariable* fv = result->addVariable("\\false");
	result->addClause(false, fv);
	Cnf::CnfVariable* top = this->ToCnf(result);
	result->addClause(true, top); // otherwise we would accept assignments that evaluate to false together with a false-assignment on the top
	return result;
}
std::vector<SatFormula*> SatFormula::CollectDisjunctions()
{
	return std::vector<SatFormula*>(1,this);
}

std::vector<Sat::SatFormula*> SatFormula::CollectConjunctions()
{
	return std::vector<SatFormula*>(1,this);
}




SatTrue::SatTrue()
:	SatFormula()
{
}
SatTrue::~SatTrue()
{
}
bool SatTrue::Evaluate(SatAssignment* assignment)
{
	return true;
}
void SatTrue::Write(std::ostream& os)
{
	os << "true";
}
Cnf::CnfVariable* SatTrue::ToCnf(Cnf::CnfFormula* formula)
{
	return formula->addVariable("\\true");
}



SatFalse::SatFalse()
:	SatFormula()
{
}
SatFalse::~SatFalse()
{
}
bool SatFalse::Evaluate(SatAssignment* assignment)
{
	return false;
}
void SatFalse::Write(std::ostream& os)
{
	os << "false";
}
Cnf::CnfVariable* SatFalse::ToCnf(Cnf::CnfFormula* formula)
{
	return formula->addVariable("\\false");
}



SatVariable::SatVariable(std::string name)
:	SatFormula()
,	name(name)
{
}
SatVariable::~SatVariable()
{
}
void SatVariable::CollectVariables(std::set<std::string>* variables)
{
	variables->insert(name);
}
bool SatVariable::Evaluate(SatAssignment* assignment)
{
	return assignment->values[name];
}
void SatVariable::Write(std::ostream& os)
{
	os << name;
}
Cnf::CnfVariable* SatVariable::ToCnf(Cnf::CnfFormula* formula)
{
	return formula->addVariable(name);
}



SatAnd::SatAnd(SatFormula* l, SatFormula* r)
:	SatFormula()
,	l(l)
,	r(r)
{
}
SatAnd::~SatAnd()
{
	delete r;
	delete l;
}
void SatAnd::CollectVariables(std::set<std::string>* variables)
{
	l->CollectVariables(variables);
	r->CollectVariables(variables);
}
bool SatAnd::Evaluate(SatAssignment* assignment)
{
	if(!l->Evaluate(assignment))
		return false;
	return r->Evaluate(assignment);
}
void SatAnd::Write(std::ostream& os)
{
	os << "(";
	l->Write(os);
	os << " && ";
	r->Write(os);
	os << ")";
}
Cnf::CnfVariable* SatAnd::ToCnf(Cnf::CnfFormula* formula)
{
	Cnf::CnfVariable* lv = l->ToCnf(formula);
	Cnf::CnfVariable* rv = r->ToCnf(formula);
	Cnf::CnfVariable* t = formula->addVariable();
	formula->addClause( true, t, false, lv, false, rv);
	formula->addClause(false, t,  true, lv);
	formula->addClause(false, t,             true, rv);
	return t;
}
std::vector<Sat::SatFormula*> SatAnd::CollectConjunctions()
{
	std::vector<Sat::SatFormula*> resultL = l->CollectConjunctions();
	std::vector<Sat::SatFormula*> resultR = r->CollectConjunctions();
	resultL.insert(resultL.end(), resultR.begin(), resultR.end());
	return resultL;
}




SatOr::SatOr(SatFormula* l, SatFormula* r)
:	SatFormula()
,	l(l)
,	r(r)
{
}
SatOr::~SatOr()
{
	delete r;
	delete l;
}
void SatOr::CollectVariables(std::set<std::string>* variables)
{
	l->CollectVariables(variables);
	r->CollectVariables(variables);
}
bool SatOr::Evaluate(SatAssignment* assignment)
{
	if(l->Evaluate(assignment))
		return true;
	return r->Evaluate(assignment);
}
void SatOr::Write(std::ostream& os)
{
	os << "(";
	l->Write(os);
	os << " || ";
	r->Write(os);
	os << ")";
}
Cnf::CnfVariable* SatOr::ToCnf(Cnf::CnfFormula* formula)
{
	Cnf::CnfVariable* lv = l->ToCnf(formula);
	Cnf::CnfVariable* rv = r->ToCnf(formula);
	Cnf::CnfVariable* t = formula->addVariable();
	formula->addClause( true, t,            false, rv);
	formula->addClause( true, t, false, lv);
	formula->addClause(false, t,  true, lv,  true, rv);
	return t;
}
std::vector<Sat::SatFormula*> SatOr::CollectDisjunctions()
{
	std::vector<Sat::SatFormula*> resultL = l->CollectDisjunctions();
	std::vector<Sat::SatFormula*> resultR = r->CollectDisjunctions();
	resultL.insert(resultL.end(), resultR.begin(), resultR.end());
	return resultL;
}



SatNot::SatNot(SatFormula* f)
:	SatFormula()
,	f(f)
{
}
SatNot::~SatNot()
{
	delete f;
}
void SatNot::CollectVariables(std::set<std::string>* variables)
{
	f->CollectVariables(variables);
}
bool SatNot::Evaluate(SatAssignment* assignment)
{
	return !f->Evaluate(assignment);
}
void SatNot::Write(std::ostream& os)
{
	os << "!";
	f->Write(os);
}
Cnf::CnfVariable* SatNot::ToCnf(Cnf::CnfFormula* formula)
{
	Cnf::CnfVariable* fv = f->ToCnf(formula);
	Cnf::CnfVariable* t = formula->addVariable();
	formula->addClause( true, t,  true, fv);
	formula->addClause(false, t, false, fv);
	return t;
}



SatIff::SatIff(SatFormula* l, SatFormula* r)
:	SatFormula()
,	l(l)
,	r(r)
{
}
SatIff::~SatIff()
{
	delete r;
	delete l;
}
void SatIff::CollectVariables(std::set<std::string>* variables)
{
	l->CollectVariables(variables);
	r->CollectVariables(variables);
}
bool SatIff::Evaluate(SatAssignment* assignment)
{
	bool lr = l->Evaluate(assignment);
	bool rr = r->Evaluate(assignment);
	return (lr && rr) || (!lr && !rr);
}
void SatIff::Write(std::ostream& os)
{
	os << "(";
	l->Write(os);
	os << " == ";
	r->Write(os);
	os << ")";
}
Cnf::CnfVariable* SatIff::ToCnf(Cnf::CnfFormula* formula)
{
	Cnf::CnfVariable* lv = l->ToCnf(formula);
	Cnf::CnfVariable* rv = r->ToCnf(formula);
	Cnf::CnfVariable* t = formula->addVariable();
	formula->addClause( true, t,  true, lv,  true, rv);
	formula->addClause( true, t, false, lv, false, rv);
	formula->addClause(false, t,  true, lv, false, rv);
	formula->addClause(false, t, false, lv,  true, rv);
	return t;
}


